#include <iostream>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>
#include <string.h>
#include <Item.h>

using namespace std;

// This program demonstrates a solution to the producer consumer 
// problem using threads, a global shared circular buffer, and 
// semaphores as the synchronization mechanism.
//
// Consider a "producer" process which is responsible for gathering 
// data about up to the minute stock prices at the stock market.
//
// Consider a "consumer" process which gets the stock prices 
// from the "producer"
//
// The original problem was taken from "Operating System Concepts: 
// 6th Edition By Silbershatz, Galvin, Gagne
//

////////////////////////////////////////////////////////////////////
// GLOBAL VARIABLES AND CONSTANTS
////////////////////////////////////////////////////////////////////

// Number of items that can be stored in the buffer
const int BUFFER_SIZE = 1000;

// Circular buffer shared by threads
Item buffer[BUFFER_SIZE];

// SEMAPHORE: Counts emtpy buffer slots
// INITIALIZE to BUFFER_SIZE because all slots are empty
sem_t empty;
const int EMPTY_INIT_VALUE=BUFFER_SIZE;

// SEMAPHORE: Counts used buffer slots
// INITIALIZE to 0 because no slots are used
sem_t used;
const int USED_INIT_VALUE=0;

////////////////////////////////////////////////////////////////////
// Function Prototypes
////////////////////////////////////////////////////////////////////

void* producer(void *);
// Precondition:  consumer pthread and semaphores initialized
// Postcondition: never returns

void* consumer(void *);
// Precondition:  consumer pthread and semaphores initialized
// Postcondition: never returns

void insertItem (const Item& theItem, int& emptySlot);
// Precondition:  call to this function must be thread safe, it 
//                accesses shared data
// Postcondition: on return, an item has been placed in the item buffer
//                and the position of the next empty slot has been updated

void removeItem (Item& theItem, int& itemSlot);
// Precondition:  call to this function must be thread safe, it 
//                accesses shared data
// Postcondition: on return an item has been removed from the item buffer
//                and the position of the next item slot has been updated

////////////////////////////////////////////////////////////////////
// Main Program
////////////////////////////////////////////////////////////////////
int main(void)
{
  // Thread initialization
  pthread_t tid0;

  // Semaphore initialization  
  sem_init(&empty,0,EMPTY_INIT_VALUE);
  sem_init(&used, 0,USED_INIT_VALUE);

  // Create thread for consumer
  pthread_create(&tid0, NULL,consumer, NULL);
  
  // Current thread is for producer
  producer(NULL);

  // Wait for thread to exit... we should never get here
  pthread_join(tid0,NULL);

  return 1;
}

////////////////////////////////////////////////////////////////////
// This thread is the PRODUCER
////////////////////////////////////////////////////////////////////
void* producer(void *)
{
  // The next empty slot in the buffer to insert item.
  int emptySlot = 0;

  while (true) {
    // Produce a new item
    Item theItem;

    // Decrement empty count, wait if it's 0, no room!
    sem_wait(&empty);            

    insertItem(theItem, emptySlot);
    
    // Increment used count, no waiting ever
    sem_post(&used);
    
  }
}

void insertItem(const Item& theItem, int & emptySlot) 
{
  // Copy item to shared buffer
  buffer[emptySlot] = theItem;

  // Circular move to next open slot in buffer
  emptySlot = (emptySlot + 1) % BUFFER_SIZE;
}

////////////////////////////////////////////////////////////////////
// This thread is the CONSUMER
////////////////////////////////////////////////////////////////////
void *consumer(void *) {
  // The next slot in the buffer to remove item.
  int itemSlot = 0;
  
  while (true) {
    // Fill this with item from producer
    Item theItem;
    
    // Decrement used count, wait if it's 0, nothing to consume!
    sem_wait(&used);
    
    removeItem(theItem, itemSlot);

    // Increment emtpy count, no waiting ever
    sem_post(&empty);

    // Output item
    cout << "consumer: " << theItem.ticker 
         << " "          << theItem.price << endl;
  }
}

void removeItem(Item& theItem, int& itemSlot) 
{
  // Copy item from shared buffer
  theItem = buffer[itemSlot];

  // Circular move to next item in buffer
  itemSlot = (itemSlot + 1) % BUFFER_SIZE;
}
